import sys
from distutils.util import strtobool
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data 
import torchvision.datasets as dsets
import torchvision.transforms as transforms
import torch.optim

class MyPerceptron(nn.Module): 
    def __init__(self, hidden_size, num_classes, sigma, device): 
        super(MyPerceptron, self).__init__()
        self.device = device
        self.hidden_size = hidden_size
        self.num_classes = num_classes
        # input:28x28, conv1 24x24x6, MaxPool1 12x12x6, Conv2 8x8x10, MaxPool2 4x4x10, FC1 160x100, FC2 100x100
        self.conv1 = nn.Conv2d(1, 6, 5)
        self.conv2 = nn.Conv2d(6, 10, 5)
        self.fc1 = nn.Linear(16*10, self.hidden_size)
        self.fc2 = nn.Linear(self.hidden_size, self.num_classes)
        self.sigma = sigma
        
    
    def forward(self, x): 
        out = F.relu(self.conv1(x)) #畳み込み層1の計算
        out = F.max_pool2d(out, (2,2)) #プーリング層1の計算
        out = F.relu(self.conv2(out)) #畳み込み層2の計算
        out = F.max_pool2d(out, (2,2))  #プーリング層2の計算
        out = out.view(-1, 16*10)
        out = self.fc1(out) #全結合層1の計算
        out = F.relu(out) 
        out = self.fc2(out) #全結合層2の計算
        out = F.relu(out)
        return out
    
    #特徴ベクトルを返すメソッド
    def getFeatureVectors(self, x):
        out = self.forward(x)
        return out
    
    #入力ベクトルを整形するメソッド
    def MyReshapeInput(self, images):
        images = images.view(-1,1,28,28)
        return images

class XShotLearningMachine():
    def __init__(self, mlp, NumOfClasses, device):
        self.templete=torch.tensor([1])
        self.Labels = torch.tensor([1])
        self.MLP = mlp
        self.IsScratch = True
        self.NumOfClasses = NumOfClasses
        self.device = device
    
    def learning(self, train_loader):
        for index, (images, labels) in enumerate(train_loader):
            images = self.MLP.MyReshapeInput(images)
            images = images.to(self.device)
            outputs = self.MLP.getFeatureVectors(images)
            if self.IsScratch:
                self.templete = outputs.clone()
                self.Labels = labels.clone()
                self.IsScratch = False
            else:
                self.templete = torch.cat([self.templete, outputs], dim=0)
                self.Labels = torch.cat([self.Labels, labels])
       
    def predictLabel(self, X):
        outputs = self.MLP.getFeatureVectors(X)
        FinalOutputs = torch.zeros([len(outputs)])
        for n in range(len(outputs)):
            each_input = outputs[n]
            each_output = torch.zeros([len(self.templete)])
            for p in range(len(self.templete)):
                each_templete = self.templete[p]
                dotproduct = torch.dot(each_templete, each_input)
                each_output[p] = dotproduct / (each_templete.norm() * each_input.norm())
            
            FinalOutputs[n] = self.Labels[(each_output.argmax()).item()]
        return FinalOutputs  
    
    #testサンプルを使った評価
    def evaluation(self, test_loader):
        totalacc = 0
        each_acc = torch.zeros([self.NumOfClasses,self.NumOfClasses])
        class_size = torch.zeros([self.NumOfClasses])
        for index, (images, labels) in enumerate(test_loader):
            images = images.to(self.device)
            images = self.MLP.MyReshapeInput(images)
            outputs = self.predictLabel(images)
            acc = (outputs==labels).sum() # 正解数を数える
            #generate recognition result matrix
            for p in range(len(labels)):
                each_acc[labels[p].item()][int(outputs[p].item())] += 1
                class_size[labels[p].item()] += 1
            totalacc += acc.item()
        return totalacc, each_acc, class_size
    
    def getLabel(self, outputs):
        return outputs.max(1)[1]

# for comparison.
class NearestNeighbourLearningMachine():
    def __init__(self, NumOfClasses, device):
        self.templete=torch.tensor([1]) 
        self.Labels = torch.tensor([1]) 
        self.IsScratch = True
        self.NumOfClasses = NumOfClasses
        self.device = device
    
    
    def learning(self, train_loader):
        for index, (images, labels) in enumerate(train_loader):
            images = images.view(-1, 28*28)
            outputs = images.clone()
            if self.IsScratch:
                self.templete = outputs.clone()
                self.Labels = labels.clone()
                self.IsScratch = False
            else:
                self.templete = torch.cat([self.templete, outputs], dim=0)
                self.Labels = torch.cat([self.Labels, labels])
                
    def predictLabel(self, X):
        FinalOutputs = torch.zeros([len(X)])
        for n in range(len(X)):
            each_input = X[n].clone()
            each_output = torch.zeros([len(self.templete)])
            for p in range(len(self.templete)):
                each_templete = self.templete[p]
                dotproduct = torch.dot(each_templete, each_input)
                
                each_output[p] = dotproduct / (each_templete.norm() * each_input.norm())
            
            FinalOutputs[n] = self.Labels[(each_output.argmax()).item()]
        return FinalOutputs  
    
    
    def evaluation(self, test_loader):
        totalacc = 0
        each_acc = torch.zeros([self.NumOfClasses,self.NumOfClasses]) 
        class_size = torch.zeros([self.NumOfClasses])
        for index, (images, labels) in enumerate(test_loader):
            images = images.view(-1, 28*28)
            outputs = self.predictLabel(images) 
            outputs = outputs.to(self.device) 
            labels = labels.to(self.device) 
            acc = (outputs==labels).sum() 
            # create recognition result matrix
            for p in range(len(labels)):
                each_acc[labels[p].item()][int(outputs[p].item())] += 1
                class_size[labels[p].item()] += 1
            totalacc += acc.item()
        return totalacc, each_acc, class_size
    
    def getLabel(self, outputs):
#         print("getLabel() outputs=%s label=%s" %(outputs, outputs.max(1)[1]) )
        return outputs.max(1)[1] 

'''#<--訂正：本プログラムではMyMNISTクラスは使用しません。class MyMNISTの定義部分は無くて良いです。
class MyMNIST(torch.utils.data.Dataset): # This class extends torch.utils.data.Dataset
    # This class generates learning samples whose labels are specified by TargetLabels[]
    def __init__(self, TargetLabels, rootPath, transform):
        self.transform = transform
        self.rootPath = rootPath
        self.dataset = dsets.MNIST(root=self.rootPath, train=True, transform=self.transform, download=True)
        self.TotalDatasetSize = 0
        self.indicies = []
        self.label = []
        for p in range(self.dataset.__len__()):
            each_input, each_label = self.dataset.__getitem__(p)
            for n in range(len(TargetLabels)):
                if each_label == TargetLabels[n]:# if label is equall to TargetLabel[n]
                    self.indicies.append(p) # save saffix into self.indicies
                    self.label.append(each_label) # label data is stored in self.label
                    self.TotalDatasetSize += 1 #count data size
        self.mydata = torch.zeros(self.TotalDatasetSize,32, 32) # init tensor
        for p in range(self.TotalDatasetSize): # generate dataset
            each_input, each_label = self.dataset.__getitem__(self.indicies[p])
            self.mydata[p] = each_input
        self.data = self.mydata.numpy()
    
    
    def __len__(self): # return dataset size
        return self.TotalDatasetSize
    
    def __getitem__(self, idx): # return each instance
        out_data = self.data[idx]
        out_label =  self.label[idx]
        if self.transform:
            out_data = self.transform(out_data) # transform out_data
        return out_data, out_label 
'''

class GetFewShotLLData(torch.utils.data.Dataset): # for testing few shot learning performance.
    # This class generates learning samples whose labels are specified by TargetLabels[]
    # Moreover, this class restricts the number of samples for each class.
    def __init__(self, TargetLabels, SizeOfEachClass, rootPath, transform):
        self.transform = transform
        self.rootPath = rootPath
        self.dataset = dsets.MNIST(root=self.rootPath, train=True, transform=self.transform, download=True)
        self.TotalDatasetSize = 0
        self.datasetsize = torch.zeros([10])
        self.indicies = []
        self.label = []
        print("GetFewShotLLData() datasetsize = %s" %(self.dataset.__len__()))
        for p in range(self.dataset.__len__()):
            each_input, each_label = self.dataset.__getitem__(p)
            for n in range(len(TargetLabels)):
                if each_label == TargetLabels[n]:
                    if self.datasetsize[TargetLabels[n]].item() < SizeOfEachClass:
                        self.indicies.append(p) 
                        self.label.append(each_label)
                        self.TotalDatasetSize += 1 
                        self.datasetsize[TargetLabels[n]] += 1
                    
        self.mydata = torch.zeros(self.TotalDatasetSize, 28, 28)
        for p in range(self.TotalDatasetSize):
            each_input, each_label = self.dataset.__getitem__(self.indicies[p])
            self.mydata[p] = each_input
        self.data = self.mydata.numpy()
    
    
    def __len__(self):
        return self.TotalDatasetSize
    
    def __getitem__(self, idx):
        out_data = self.data[idx]
        out_label =  self.label[idx]
        if self.transform:
            out_data = self.transform(out_data)
        return out_data, out_label 

class main():     
    def __init__(self, InitialLearning, parameterfilename, Iterations, Prototypes, NumDivision):
        print(" main() _init_() started!")
        self.input_size = 784
        self.hidden_size = 100
        self.num_classes = 50
        self.batch_size = 50
        self.NumberOfIterations = Iterations
        self.NumberOfPrototypes = Prototypes
        self.num_division = NumDivision 
        self.label_step = int(10/self.num_division)
        print("InitialLearning=%s, parameterfilename=%s, Iterations=%s, # of Prototypes =%s, # of divisions=%s" \
        %(InitialLearning, parameterfilename, self.NumberOfIterations, self.NumberOfPrototypes, self.num_division))

        # データセット        
        self.test_dataset = dsets.MNIST(root='./data', train=False, transform=transforms.ToTensor(), download=True)
        test_loader = torch.utils.data.DataLoader(dataset=self.test_dataset, 
                                                  batch_size=self.batch_size, shuffle=True,pin_memory=False)
        self.train_dataset = dsets.MNIST(root='./data', train=True, transform=transforms.ToTensor())

        # Choose GPU or CPU automatically.
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        
        # Generate MyPerceptron object.
        self.model = MyPerceptron(self.hidden_size, self.num_classes,1.0, self.device)
        print("network = %s" %(self.model))
        if InitialLearning==False: # If it is not the initial learning, it reads parameters from a specified file.
            print("Parameter file is loaded")
            self.model.load_state_dict(torch.load(parameterfilename))
        # self.model -> device
        self.model = self.model.to(self.device)
        self.optimizer = torch.optim.Adagrad(self.model.parameters(True), lr=0.01)
        if InitialLearning==True: #initial learning mode
            print("Initial Learning is started!")
            # Siamese learning
            for i in range(self.NumberOfIterations):
                train_loader = torch.utils.data.DataLoader(dataset=self.train_dataset, batch_size=10, shuffle=True,pin_memory=False)
                totalloss = self.SiameseLearning(self.model, train_loader)
                print("Siamese learning : %s : err=%s" %(i, totalloss))
             #save resultant parameters for the Siamese neural network.
            torch.save(self.model.state_dict(), parameterfilename)
        
        xshot = XShotLearningMachine(self.model, 10, self.device)
        NN= NearestNeighbourLearningMachine(10, self.device)
        # Few shot learning data
        for n in range(0, self.num_division):
            ##このラウンドで学習させるラベルをtarget_labelsにセットする ##
            targetLabels = []
            for l in range(n*self.label_step, (n+1)*self.label_step):
                targetLabels.append(l)
            ##target_labels[]に従って学習用データをロードする ##
            fewShotllData = GetFewShotLLData(targetLabels, self.NumberOfPrototypes, rootPath='./data', transform=transforms.ToTensor())
            train_loader = torch.utils.data.DataLoader(dataset=fewShotllData, batch_size=100, shuffle=True)
            
            xshot.learning(train_loader)
            totalacc, AccMatrix, class_size = xshot.evaluation(test_loader)
            print("----- XShotLL using Siamese Convolutional Network -----")
            for c in range(10):
                print("%s : %s " %(c, (float)(AccMatrix[c][c].item()/class_size[c].item())))
            
            print("----- NearestNeighbour -----")   
            NN.learning(train_loader)
            totalacc, AccMatrix, class_size = NN.evaluation(test_loader)
            for c in range(10):
                print("%s : %s " %(c, (float)(AccMatrix[c][c].item()/class_size[c].item())))
            
    def SiameseLearning(self, model, train_loader):
        totalloss = 0
        for index, (images, labels) in enumerate(train_loader):
            images = images.to(self.device)
            labels = labels.to(self.device)
            images = model.MyReshapeInput(images)
            self.optimizer.zero_grad() 
            outputs = model.getFeatureVectors(images) # get output vectors
            # make learning sample
            featureUpShift, featureDownShift, UpShiftmatchflugs, DownShiftmatchflugs = self.createChunkData(outputs, labels)
    
            loss = self.ContrastiveLoss(outputs, featureDownShift, DownShiftmatchflugs) # calculate the loss for DownShifted feature vectors
            totalloss += loss.item()
            loss.backward(retain_graph=True) # make sure that the backward is repeated two times.
            # To do this, retain_graph=True should be set.
            loss = self.ContrastiveLoss(outputs, featureUpShift, UpShiftmatchflugs) # calculate the loss for UpShifted feature vectors.
            totalloss += loss.item()
            loss.backward()
            self.optimizer.step()
        return totalloss
        
        
        
    def createChunkData(self, featureTensor, labelTensor): # create learning samples for the Siamese neural networks
        labelUpShift = torch.roll(labelTensor, -1, dims=0).detach()
        lableDownShift = torch.roll(labelTensor, 1, dims=0).detach()
        featureUpshift = torch.roll(featureTensor, -1, dims=0).detach()
        featureDownShift = torch.roll(featureTensor, 1, dims=0).detach()
        ones = torch.ones(labelTensor.size()[0])
        zeros = torch.zeros(labelTensor.size()[0])
        ones = ones.to(self.device)
        zeros = zeros.to(self.device)
        UpShiftmatchflugs = torch.where(labelTensor==labelUpShift, ones, zeros)
        DownShiftmatchflugs = torch.where(labelTensor==lableDownShift, ones, zeros)
        return featureUpshift, featureDownShift, UpShiftmatchflugs, DownShiftmatchflugs

    def ContrastiveLoss(self, RealFeatureVectors, TargetFeatureVectors, MatchFlags):
        length = len(RealFeatureVectors)
        diff = TargetFeatureVectors - RealFeatureVectors
        sqr_distance =  torch.mul(diff,diff)
        sqr_distance = sqr_distance.sum(1).view(-1,1)
        distance = torch.sqrt(sqr_distance)
        distance = distance.to(self.device)
        sqr_distance = sqr_distance.to(self.device)
        result = torch.tensor([0.])
        MatchFlags = MatchFlags.view(-1,length,1)
        result = 0.5 *( torch.mul(MatchFlags, sqr_distance) + (1-MatchFlags)* (F.relu(1-distance) ** 2))  
        result = result.view(-1,length) # vertical vector -> horizontal vector
        result = result.sum()
        return result.clone()

if __name__ == "__main__":
    if len(sys.argv) < 5:
        print("Usage python3 fewshot.py [IntialLearning:True / False][parameter filename][Iterations][Prototypes][NumDivision]")
    else:
        print("argv[1]=%s bool=%s %s " %(sys.argv[1],strtobool(sys.argv[1]),len(sys.argv)))
        MAIN = main(strtobool(sys.argv[1]), sys.argv[2], int(sys.argv[3]), int(sys.argv[4]), int(sys.argv[5]))
